@@ -21,6 +21,7 @@ from cvnn.layers import ComplexDense, ComplexDropout
 from cvnn.utils import transform_to_real, randomize, transform_to_real_map_function
 from cvnn.real_equiv_tools import get_real_equivalent
 from cvnn.utils import median_error
+from cvnn.initializers import ComplexGlorotUniform
 # typing
 from pathlib import Path
 from typing import Union, Optional, List, Tuple
@@ -42,16 +43,15 @@ class MonteCarlo:
         self.models = []
         self.pandas_full_data = pd.DataFrame()
         self.monte_carlo_analyzer = MonteCarloAnalyzer()  # All at None
+        self.verbose = 1
         self.output_config = {
             'plot_all': False,
             'confusion_matrix': False,
             'excel_summary': True,
-            'debug': False,
             'summary_of_run': True,
             'tensorboard': False,
             'save_weights': False,
-            'safety_checkpoints': False,
-            'early_stop': False
+            'safety_checkpoints': False
         }
 
     def add_model(self, model: Type[Model]):
@@ -60,14 +60,35 @@ class MonteCarlo:
         """
         self.models.append(model)
 
+    @staticmethod
+    def _parse_verbose(verbose : Union[str, int, bool]) -> int:
+        if isintance(verbose, bool):
+            verbose = 2 if verbose else 1
+        elif isinstance(verbose, str):
+            if verbose.lower() == 'silent':
+                verbose = 0
+            elif verbose.lower() == 'debug':
+                verbose = 2
+            else:
+                raise ValueError(f"Unknown verbose mode {verbose}")
+        else:
+            try:
+                verbose = int(verbose)
+                if verbose > 2 or verbose < 0:
+                    raise ValueError(f"verbose should be one of 0, 1 or 2, received {verbose}")
+            except e:
+                raise ValueError(f"Cannot cast verbose = {verbose} to int")
+        return verbose
+
     def run(self, x, y, data_summary: str = '',
             real_cast_modes: Optional[Union[str, List[Optional[str]], Tuple[Optional[str]]]] = None,
             validation_split: float = 0.2,
             validation_data: Optional[Union[Tuple[np.ndarray, np.ndarray], data.Dataset]] = None,
             # TODO: Add the tuple of validation data details.
             test_data: Optional[Union[Tuple[np.ndarray, np.ndarray], data.Dataset]] = None,
-            iterations: int = 100, epochs: int = 10, batch_size: int = 100,
-            shuffle: bool = True, debug: bool = False, display_freq: int = 1):
+            iterations: int = 100, epochs: int = 10, batch_size: int = 100, early_stop: bool = False,
+            shuffle: bool = True, verbose: Optional[Union[bool, int, str]] = 1, display_freq: int = 1,
+            same_weights: bool = False, process_dataset: bool = True):
         """
         This function is used to compare all models added with `self.add_model` method.
         Runs the iteration dataset (x, y).
@@ -114,11 +135,17 @@ class MonteCarlo:
         :param display_freq: Integer (Default 1). Only relevant if validation data is provided.
             Frequency on terms of epochs before running the validation.
         :param shuffle: (Boolean) Whether to shuffle the training data before each epoch.
-        :param debug:
+        :param verbose: Different modes according to number:
+            - 0 or 'silent': No output at all
+            - 1 or False: Progress bar per iteration
+            - 2 or True or 'debug': Progress bar per epoch
+        :param early_stop: (Default: False) Wheather to implement early stop on training.
+        :param same_weights: (Default False) If True it will use the same weights at each iteration.
         :return: (string) Full path to the run_data.csv generated file.
             It can be used by cvnn.data_analysis.SeveralMonteCarloComparison to compare several runs.
         """
-        self.output_config['debug'] = debug
+        if verbose:
+            self.verbose = self._parse_verbose(verbose)
         test_data_cols = None
         if test_data is not None:
             test_data_cols = ['network'] + [n.get_config()['name'] for n in self.models[0].metrics]
@@ -130,13 +157,25 @@ class MonteCarlo:
             w_save.append(model.get_weights())  # Save model weight
         # np.save(self.monte_carlo_analyzer.path / "initial_debug_weights.npy", np.array(w_save))     # TODO
         for it in range(iterations):
-            if debug:
+            if self.verbose == 2:
                 logger.info("Iteration {}/{}".format(it + 1, iterations))
             for i, model in enumerate(self.models):
                 x_fit, val_data_fit, test_data_fit = self._get_fit_dataset(model.inputs[0].dtype.is_complex, x,
                                                                            validation_data, test_data,
-                                                                           real_cast_modes[i])
-                model.set_weights(w_save[i])
+                                                                           real_cast_modes[i],
+                                                                           process_dataset=process_dataset)
+                clone_model = tf.keras.models.clone_model(model)
+                if isinstance(model.loss, tf.keras.losses.Loss):
+                    loss = model.loss.__class__.from_config(config=model.loss.get_config())
+                else:
+                    loss = model.loss
+                clone_model.compile(optimizer=model.optimizer.__class__.from_config(model.optimizer.get_config()),
+                                    loss=loss,
+                                    metrics=['accuracy'])   # TODO: Until the issue is solved, I need to force metrics
+                # https://github.com/tensorflow/tensorflow/issues/40030
+                # https://stackoverflow.com/questions/62116136/tensorflow-keras-metrics-not-showing/69193373
+                if same_weights:
+                    clone_model.set_weights(w_save[i])
                 temp_path = self.monte_carlo_analyzer.path / f"run/iteration{it}_model{i}_{model.name}"
                 os.makedirs(temp_path, exist_ok=True)
                 callbacks = []
@@ -144,18 +183,19 @@ class MonteCarlo:
                     tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir=temp_path / 'tensorboard',
                                                                           histogram_freq=1)
                     callbacks.append(tensorboard_callback)
-                if self.output_config['early_stop']:
+                if early_stop:
                     eas = tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=10)
                     callbacks.append(eas)
-                run_result = model.fit(x_fit, y, validation_split=validation_split, validation_data=val_data_fit,
-                                       epochs=epochs, batch_size=batch_size,
-                                       verbose=debug, validation_freq=display_freq,
-                                       callbacks=callbacks, shuffle=shuffle)
-                test_results = self._inner_callback(model, validation_data, confusion_matrix, real_cast_modes[i], i,
-                                                    run_result, test_results, test_data_fit, temp_path)
+                run_result = clone_model.fit(x_fit, y, validation_split=validation_split, validation_data=val_data_fit,
+                                             epochs=epochs, batch_size=batch_size,
+                                             verbose=self.verbose==2, validation_freq=display_freq,
+                                             callbacks=callbacks, shuffle=shuffle)
+                test_results = self._inner_callback(clone_model, validation_data, confusion_matrix, real_cast_modes[i],
+                                                    i, run_result, test_results, test_data_fit, temp_path)
             self._outer_callback(pbar)
         return self._end_callback(x, y, iterations, data_summary, real_cast_modes, epochs, batch_size,
                                   confusion_matrix, test_results, pbar, w_save)
+        # TODO: What was the idea of save_weights? Is it necessary or it was only debugging?
 
     def _check_real_cast_modes(self, real_cast_modes):
         # TODO: I can check the real models input size corresponds to the real_cast_mode. And change it with a warning?
@@ -185,7 +225,9 @@ class MonteCarlo:
                 val_data_fit = (transform_to_real(validation_data[0], mode=polar), validation_data[1])
         return val_data_fit
 
-    def _get_fit_dataset(self, is_complex: bool, x, validation_data, test_data, polar):
+    def _get_fit_dataset(self, is_complex: bool, x, validation_data, test_data, polar, process_dataset):
+        if not process_dataset:
+            return x, validation_data, test_data
         if isinstance(x, tf.data.Dataset):
             if not is_complex:
                 x_fit = x.map(lambda imag, label: transform_to_real_map_function(imag, label, mode=polar))
@@ -209,7 +251,7 @@ class MonteCarlo:
         pbar = None
         # Reset data frame
         self.pandas_full_data = pd.DataFrame()
-        if not self.output_config['debug']:
+        if self.verbose == 1:
             pbar = tqdm(total=iterations)
         if self.output_config['confusion_matrix']:
             confusion_matrix = []
@@ -225,7 +267,7 @@ class MonteCarlo:
 
     def _end_callback(self, x, y, iterations, data_summary, polar, epochs, batch_size,
                       confusion_matrix, test_results, pbar, w_save):
-        if not self.output_config['debug']:
+        if self.verbose == 1:
             pbar.close()
         self.pandas_full_data = self.pandas_full_data.reset_index(drop=True)
         self.monte_carlo_analyzer.set_df(self.pandas_full_data)
@@ -292,7 +334,7 @@ class MonteCarlo:
         return test_results
 
     def _outer_callback(self, pbar):
-        if not self.output_config['debug']:
+        if self.verbose == 1:
             pbar.update()
         if self.output_config['safety_checkpoints']:
             # Save checkpoint in case Monte Carlo stops in the middle
@@ -344,9 +386,9 @@ class MonteCarlo:
         for i, model in enumerate(self.models):
             json_dict[str(i)] = {
                 'name': model.name,
-                'loss': tf.keras.losses.serialize(model.loss),
-                'optimizer': tf.keras.optimizers.serialize(model.optimizer),
-                'layers': [tf.keras.layers.serialize(layer) for layer in model.layers]
+                'loss': model.loss if isinstance(model.loss, str) else model.loss.get_config(),  # Not yet support function loss
+                'optimizer': model.optimizer.get_config(),
+                'layers': [layer.get_config() for layer in model.layers]
             }
         with open(self.monte_carlo_analyzer.path / 'models_details.json', 'w') as fp:
             json.dump(str(json_dict), fp)
@@ -430,11 +472,11 @@ def run_gaussian_dataset_montecarlo(iterations: int = 30, m: int = 10000, n: int
                                     epochs: int = 300, batch_size: int = 100, display_freq: int = 1,
                                     optimizer='sgd', validation_split: float = 0.2,  # TODO: Add typing here
                                     shape_raw: List[int] = None, activation: t_activation = 'cart_relu',
-                                    debug: bool = False, do_all: bool = True, tensorboard: bool = False,
+                                    verbose: bool = False, do_all: bool = True, tensorboard: bool = False,
                                     polar: Optional[Union[str, List[Optional[str]], Tuple[Optional[str]]]] = None,
                                     capacity_equivalent: bool = True, equiv_technique: str = 'ratio',
                                     dropout: Optional[float] = None, models: Optional[List[Model]] = None,
-                                    plot_data: bool = True, early_stop: bool = False) -> str:
+                                    plot_data: bool = True, early_stop: bool = False, shuffle: bool = True) -> str:
     """
     This function is used to compare CVNN vs RVNN performance over statistical non-circular data.
         1. Generates a complex-valued gaussian correlated noise with the characteristics given by the inputs.
@@ -470,7 +512,10 @@ def run_gaussian_dataset_montecarlo(iterations: int = 30, m: int = 10000, n: int
         For example [64] will generate a CVNN with one hidden layer of size 64.
         Default None will default to example.
     :param activation: Activation function to be used at each hidden layer
-    :param debug:
+    :param verbose: Different modes according to number:
+            - 0 or 'silent': No output at all
+            - 1 or False: Progress bar per iteration
+            - 2 or True or 'debug': Progress bar per epoch
     :param tensorboard: If True, it will generate tensorboard outputs to check training values.
     :param polar: Boolean weather the RVNN should receive real and imaginary part (False) or amplitude and phase (True)
     :param do_all: If true (default) it creates a `plot/` folder with the plots generated by MonteCarloAnalyzer.do_all()
@@ -491,13 +536,13 @@ def run_gaussian_dataset_montecarlo(iterations: int = 30, m: int = 10000, n: int
         return run_montecarlo(models=models, dataset=dataset, open_dataset=None,
                               iterations=iterations, epochs=epochs, batch_size=batch_size, display_freq=display_freq,
                               validation_split=validation_split, validation_data=None,
-                              debug=debug, polar=polar, do_all=do_all, tensorboard=tensorboard, do_conf_mat=False,
-                              plot_data=plot_data, early_stop=early_stop)
+                              verbose=verbose, polar=polar, do_all=do_all, tensorboard=tensorboard, do_conf_mat=False,
+                              plot_data=plot_data, early_stop=early_stop, shuffle=shuffle)
     else:
         return mlp_run_real_comparison_montecarlo(dataset=dataset, open_dataset=None, iterations=iterations,
                                                   epochs=epochs, batch_size=batch_size, display_freq=display_freq,
                                                   optimizer=optimizer, shape_raw=shape_raw, activation=activation,
-                                                  debug=debug, polar=polar, do_all=do_all,
+                                                  verbose=verbose, polar=polar, do_all=do_all,
                                                   tensorboard=tensorboard,
                                                   capacity_equivalent=capacity_equivalent,
                                                   equiv_technique=equiv_technique,
@@ -511,10 +556,11 @@ def run_montecarlo(models: List[Model], dataset: cvnn.dataset.Dataset, open_data
                    validation_split: float = 0.2,
                    validation_data: Optional[Union[Tuple, data.Dataset]] = None,
                    # TODO: Add vallidation data tuple details
-                   debug: Union[bool, int] = False, do_conf_mat: bool = False, do_all: bool = True,
+                   verbose: Union[bool, int] = False, do_conf_mat: bool = False, do_all: bool = True,
                    tensorboard: bool = False,
                    polar: Optional[Union[str, List[Optional[str]], Tuple[Optional[str]]]] = None,
-                   plot_data: bool = False, early_stop: bool = True) -> str:
+                   plot_data: bool = False, early_stop: bool = False, shuffle: bool = True,
+                   preprocess_data: bool = True) -> str:
     """
     This function is used to compare different neural networks performance.
     1. Runs simulation and compares them.
@@ -526,13 +572,16 @@ def run_montecarlo(models: List[Model], dataset: cvnn.dataset.Dataset, open_data
 
     :param models: List of cvnn.CvnnModel to be compared.
     :param dataset: cvnn.dataset.Dataset with the dataset to be used on the training
-    :param open_dataset: (None)
+    :param open_dataset: (Default: None)
         If dataset is saved inside a folder and must be opened, path of the Dataset to be opened. Else None (default)
     :param iterations: Number of iterations to be done for each model
     :param epochs: Number of epochs for each iteration
     :param batch_size: Batch size at each iteration
     :param display_freq: Frequency in terms of epochs of when to do a checkpoint.
-    :param debug:
+    :param verbose: Different modes according to number:
+            - 0 or 'silent': No output at all
+            - 1 or False: Progress bar per iteration
+            - 2 or True or 'debug': Progress bar per epoch
     :param polar: Boolean weather the RVNN should receive real and imaginary part (False) or amplitude and phase (True)
     :param do_all: If true (default) it creates a `plot/` folder with the plots generated by MonteCarloAnalyzer.do_all()
     :param validation_split: Float between 0 and 1.
@@ -556,30 +605,39 @@ def run_montecarlo(models: List[Model], dataset: cvnn.dataset.Dataset, open_data
     for model in models:
         # model.training_param_summary()
         monte_carlo.add_model(model)
-    if not open_dataset:
+    if not open_dataset and isinstance(dataset, dp.Dataset):
         dataset.save_data(monte_carlo.monte_carlo_analyzer.path)
     monte_carlo.output_config['excel_summary'] = False
     monte_carlo.output_config['tensorboard'] = tensorboard
     monte_carlo.output_config['confusion_matrix'] = do_conf_mat
     monte_carlo.output_config['plot_all'] = do_all
-    monte_carlo.output_config['early_stop'] = early_stop
-    if plot_data:
+    if plot_data and isinstance(dataset, dp.Dataset):
         dataset.plot_data(overlapped=True, showfig=False, save_path=monte_carlo.monte_carlo_analyzer.path,
                           library='matplotlib')
-    monte_carlo.run(dataset.x, dataset.y, iterations=iterations,
+    if isinstance(dataset, dp.Dataset):
+        x = dataset.x
+        y = dataset.y
+        data_summary = dataset.summary()
+    else:
+        x = dataset
+        y = None
+        data_summary = ""
+    monte_carlo.run(x, y, iterations=iterations,
                     validation_split=validation_split, validation_data=validation_data,
-                    epochs=epochs, batch_size=batch_size, display_freq=display_freq,
-                    shuffle=False, debug=debug, data_summary=dataset.summary(), real_cast_modes=polar)
+                    epochs=epochs, batch_size=batch_size, display_freq=display_freq, early_stop=early_stop,
+                    shuffle=shuffle, verbose=verbose, data_summary=data_summary, real_cast_modes=polar,
+                    process_dataset=preprocess_data)
 
     # Save data to remember later what I did.
     _save_montecarlo_log(iterations=iterations,
                          path=str(monte_carlo.monte_carlo_analyzer.path),
                          models_names=[str(model.name) for model in models],
-                         dataset_name=dataset.dataset_name,
-                         num_classes=str(dataset.y.shape[1]),
+                         dataset_name=data_summary,
+                         num_classes=str(dataset.y.shape[1]) if isinstance(dataset, dp.Dataset) else "",    # TODO: GET THIS
                          polar_mode=str(polar),
-                         dataset_size=str(dataset.x.shape[0]),
-                         features_size=str(dataset.x.shape[1]), epochs=epochs, batch_size=batch_size
+                         dataset_size=str(dataset.x.shape[0]) if isinstance(dataset, dp.Dataset) else "",
+                         features_size=str(dataset.x.shape[1]) if isinstance(dataset, dp.Dataset) else "",
+                         epochs=epochs, batch_size=batch_size
                          # filename='./log/run_data.csv'
                          )
     return str("./log/run_data.csv")
@@ -591,13 +649,13 @@ def mlp_run_real_comparison_montecarlo(dataset: cvnn.dataset.Dataset, open_datas
                                        optimizer='adam',  # TODO: Typing
                                        shape_raw=None, activation: t_activation = 'cart_relu',
                                        output_activation: t_activation = DEFAULT_OUTPUT_ACT,
-                                       debug: Union[bool, int] = False, do_all: bool = True,
+                                       verbose: Union[bool, int] = False, do_all: bool = True,
                                        polar: Optional[Union[str, List[Optional[str]], Tuple[Optional[str]]]] = None,
                                        dropout: float = 0.5, validation_split: float = 0.2,
                                        validation_data: Optional[Union[Tuple, data.Dataset]] = None,
                                        # TODO: Add typing of tuple
                                        capacity_equivalent: bool = True, equiv_technique: str = 'ratio',
-                                       shuffle: bool = False, tensorboard: bool = False, do_conf_mat: bool = False,
+                                       shuffle: bool = True, tensorboard: bool = False, do_conf_mat: bool = False,
                                        plot_data: bool = True) -> str:
     """
     This function is used to compare CVNN vs RVNN performance over any dataset.
@@ -623,7 +681,10 @@ def mlp_run_real_comparison_montecarlo(dataset: cvnn.dataset.Dataset, open_datas
         For example [64] will generate a CVNN with one hidden layer of size 64.
         Default None will default to example.
     :param activation: Activation function to be used at each hidden layer
-    :param debug:
+    :param verbose: Different modes according to number:
+            - 0 or 'silent': No output at all
+            - 1 or False: Progress bar per iteration
+            - 2 or True or 'debug': Progress bar per epoch
     :param polar: Boolean weather the RVNN should receive real and imaginary part (False) or amplitude and phase (True)
     :param do_all: If true (default) it creates a `plot/` folder with the plots generated by MonteCarloAnalyzer.do_all()
     :param dropout: (float) Dropout to be used at each hidden layer. If None it will not use any dropout.
@@ -673,7 +734,7 @@ def mlp_run_real_comparison_montecarlo(dataset: cvnn.dataset.Dataset, open_datas
     sleep(1)  # I have error if not because not enough time passed since creation of models to be in diff folders
     monte_carlo.run(dataset.x, dataset.y, iterations=iterations,
                     epochs=epochs, batch_size=batch_size, display_freq=display_freq,
-                    shuffle=shuffle, debug=debug, data_summary=dataset.summary(), real_cast_modes=polar,
+                    shuffle=shuffle, verbose=verbose, data_summary=dataset.summary(), real_cast_modes=polar,
                     validation_split=validation_split, validation_data=validation_data)
 
     # Save data to remember later what I did.
@@ -741,7 +802,7 @@ def get_mlp(input_size, output_size,
         )
     else:  # len(shape_raw) > 0:
         for s in shape_raw:
-            shape.append(ComplexDense(units=s, activation=activation))  # Add dropout!
+            shape.append(ComplexDense(units=s, activation=activation))
             if dropout is not None:
                 shape.append(ComplexDropout(rate=dropout))
         shape.append(ComplexDense(units=output_size, activation=output_activation))
